grammar CShell;
options {output=AST; ASTLabelType=CShellAST;backtrack=true;}
  
tokens {
	ARGS; FIELDS; BLOCK; CALL;FOR1;FOR2;FOR3;PREV;SUFFIX; INC; DEC; ELSE; ELSEIF;ARRAY;ARRAY_ELEMENT;
	IF='if'; ASSIGN='='; PRINT='print';
	FOR='for'; RETURN='return'; LENGTH='length'; DEF='def'; ADD='+'; SUB='-';
	MUL='*'; DIV='/';EQ='=='; LT='<';GT='>';
	}
@members {
    Interpreter interp;
    Scope currentScope;
    public CShellParser(TokenStream input, Interpreter interp) {
        this(input);
        this.interp = interp;
        currentScope = interp.globalScope;
    }
}
@header
{
package cshell.interpreter;
}
@lexer::header
{
package cshell.interpreter;
}

program
	:	(function_definition |statement)* EOF 
		-> ^(BLOCK statement*);

statement
	:	assign^ ';'				
	|   inc_dec^';'            
	|	'return' expr ';'  			-> ^('return' expr)
	|	'print' '(' expr ')' ';'	 			-> ^('print' expr)
	|	'if' '(' expr ')' slist elseif_statement_all  else_statement? 
	    -> ^('if' expr slist elseif_statement_all ^(ELSE else_statement?))
	|	for_statement
	|   array_statement_init
	|   call ';'                    ->call
	|	NL							->
	;  

function_definition @init{FunctionSymbol fs = null;}
	:	'def' ID 
		{
        fs = new FunctionSymbol($ID.text,currentScope);
        currentScope.define(fs); // def method in globals
        currentScope = fs;       // set current scope to method scope
        }
        '(' (vardef (',' vardef)* )? ')'
		{/*currentScope = new LocalScope(fs);*/}  
		slist
		{
		fs.blockAST = $slist.tree;
		//System.out.println(currentScope.getScopeName());
		currentScope = currentScope.getEnclosingScope();  //  ָFunctionSymbol
		//  meָGlobal Scope
		
	    //System.out.println(currentScope.getScopeName());
		}
		-> // pass nothing to interpreter
	;
vardef @init{VariableSymbol vs = null;}
	:	ID
		{
		$ID.tree.scope = currentScope;
		vs = new VariableSymbol($ID.text);
		currentScope.define(vs);
		}
	;
call
@after {
	$call.tree.scope = currentScope;
	//$call.tree.symbol = currentScope.resolve($name.text);
}
	:	ID '(' (expr (',' expr )*)? ')' -> ^(CALL ID expr*) ;
	
else_statement:'else' slist-> slist;
elseif_statement_all: ('else' 'if' '(' expr ')'  slist)* ->^(ELSEIF ^(ELSEIF expr slist)*);

for_statement
	:	
	'for' '(' {currentScope = new LocalScope(currentScope);
	            
	          }
	          
	          for1?
	          
	          ';' for2?
	          ';' for3? ')' slist
	          {currentScope = currentScope.getEnclosingScope();}			
	      -> ^('for'  ^(FOR1 for1?) ^(FOR2 for2?) ^(FOR3 for3?)  slist)
	;	

for1	    :  assign(',' assign)* ->assign+;
for2:	expr;
for3:	expr(',' expr)* -> expr+;
// arr = [1,2,3,4,5];
array_statement_init:
      ID '=' '[' array_values? ']'';' 
      ->^(ARRAY ID array_values?)
;
array_value:INT|FLOAT|CHAR|STRING|expr;
array_values:array_value(',' array_value)* -> array_value+;

//array_index:INT^|ID^;
array_element:a = ID'[' b = expr ']' ->^(ARRAY_ELEMENT $a $b);

assign_id:ID^|array_element; 

assign: (assign_id '=' expr)-> ^('=' assign_id expr);


slist   // ԭݣ ':' NL (statement)+ '.' NL
	:	NL*'{' NL* (statement)* NL* '}'NL* 	-> ^(BLOCK statement*)

	;     
length_fun:'length' '(' assign_id ')'      -> ^('length' assign_id) ;
expr:	inc_dec|assign|(addexpr (('=='|'<'|'>')^ addexpr)?)
        |length_fun ;
inc_dec_id:ID^|array_element^;
inc_dec: ('++'  inc_dec_id)->^(PREV INC inc_dec_id)
        |('--'inc_dec_id)->^(PREV DEC inc_dec_id)
        |(inc_dec_id'++')->^(SUFFIX INC inc_dec_id)
        |(inc_dec_id'--')->^(SUFFIX DEC inc_dec_id);

addexpr  // mulexpr (('+'|'-')^ mulexpr)*
	:	 mulexpr (('+'|'-')^ mulexpr)*  
	;
 
mulexpr 
	:	atom (('*'|'/')^ atom)*
	;

atom   
	:	INT		    
	|	CHAR	    
	|	FLOAT	    
	|	STRING	
	|   array_element 
	|	ID	
	|   call	    
	|	'(' expr ')' -> expr
	;
  
// L e x i c a l  R u l e s
NL	:	'\r'? '\n' ;
ID  :   '$'?(LETTER|'_') (LETTER | '0'..'9')*  ;

fragment

LETTER
	:   ('a'..'z' | 'A'..'Z')
    ;

    
CHAR:	'\'' . '\'' ;

STRING:	'\"' .* '\"' ;

INT :   '-'?'0'..'9'+ ;
    
FLOAT
	:	INT '.' INT*
	|	'.' INT+
	;

WS  :   (' '|'\t') {$channel=HIDDEN;} ;

SL_COMMENT
    :   '//' ~('\r'|'\n')* {$channel=HIDDEN;}
    |  '/*' .* '*/'{$channel=HIDDEN;}
    ;		