/*
 * ߣ
 * 
 * ʹ
 * 
 * ʱ䣺ĩպĵһ
 * 
 */
package cshell.interpreter;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

import org.antlr.runtime.ANTLRInputStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.Token;
import org.antlr.runtime.TokenRewriteStream;




public class Interpreter
{
	public InterpreterListener listener = // default response to messages
	new InterpreterListener()
	{
		public void info(String msg)
		{
			System.out.println(msg);
		}

		public void error(String msg)
		{
			System.err.println(msg);
		}

		public void error(String msg, Exception e)
		{
			error(msg);
			e.printStackTrace(System.err);
		}

		public void error(String msg, Token t)
		{
			error("line " + t.getLine() + ": " + msg);
		}
	};
	public static final ReturnValue sharedReturnValue = new ReturnValue();
	GlobalScope globalScope; // global scope is filled by the parser
	ScopeMemorySpace globals = new ScopeMemorySpace("globals"); // global memory
	ScopeMemorySpace currentSpace = globals;
	Stack<ScopeMemorySpace> stack = new Stack<ScopeMemorySpace>();

	public void process(InputStream input) throws RecognitionException,
			IOException
	{

		globalScope = new GlobalScope();
		CShellLexer lexer = new CShellLexer(new ANTLRInputStream(input));

		TokenRewriteStream tokens = new TokenRewriteStream(lexer);
		CShellParser parser = new CShellParser(tokens, this);
		parser.setTreeAdaptor(new CShellAdapter());

		CShellParser.program_return r = parser.program();
		if (parser.getNumberOfSyntaxErrors() == 0)
		{
			CShellAST root = (CShellAST) r.getTree();

			block(root);
		}
	}

	/** visitor dispatch according to node token type */

	public Object exec(CShellAST ast)
	{
		try
		{
			switch (ast.getType())
			{
				case CShellParser.BLOCK:
					block(ast);
					break;
				case CShellParser.ASSIGN:
					assign(ast);
					break;
				case CShellParser.LENGTH:
					return length(ast);
				case CShellParser.ARRAY:
					arrayStat(ast);
					break;

				case CShellParser.RETURN:
					ret(ast);
					break;
				case CShellParser.PRINT:
					print(ast);
					break;
				case CShellParser.IF:
					ifstat(ast);
					break;
				case CShellParser.FOR:
					forloop(ast);
					break;
				case CShellParser.CALL:
					return call(ast);

				case CShellParser.ADD:
					return add(ast);
				case CShellParser.PREV:
				case CShellParser.SUFFIX:
					return incAndDec(ast);

				case CShellParser.SUB:
					return op(ast);
				case CShellParser.MUL:
				case CShellParser.DIV:
					return op(ast);
				case CShellParser.EQ:
					return eq(ast);
				case CShellParser.LT:
					return lt(ast);
				case CShellParser.GT:
					return gt(ast);
				case CShellParser.INT:
					return Integer.parseInt(ast.getText());
				case CShellParser.CHAR:
					return new Character(ast.getText().charAt(1));
				case CShellParser.FLOAT:
					return Float.parseFloat(ast.getText());
				case CShellParser.STRING:
					String s = ast.getText();
					return s.substring(1, s.length() - 1);
				case CShellParser.ID:
				case CShellParser.ARRAY_ELEMENT:
					return load(ast);
				default: // catch unhandled node types
					throw new UnsupportedOperationException("޷"
							+ ast.getText() + "<" + ast.getType() + ">");
			}
		}
		catch (Exception e)
		{
			listener.error("쳣ԭ " + ast.toStringTree(), e);
		}
		return null;
	}

	private void assign(CShellAST ast)
	{

		CShellAST id = (CShellAST) ast.getChild(0);
		if (id.getType() == CShellParser.ARRAY_ELEMENT)
		{

			id = (CShellAST) id.getChild(0);
			ScopeMemorySpace space = getScopeMemorySpace(id.getText());
			Object obj = space.get(id.getText());
			if (obj == null)
			{
				listener.error("<" + id.getText() + ">");
				return;
			}
			List array = (List) obj;
			obj =  exec((CShellAST) ast.getChild(0).getChild(1));
			if(obj == null)
			{
				return;
			}
			int index = Integer.parseInt(String.valueOf(obj));
			if (index >= array.size())
			{
				listener.error("Χ<" + index + ">");
				return;
			}
			obj = exec((CShellAST) ast.getChild(1));

			array.set(index, obj);

		}
		else
		{
			// ȡID

			CShellAST expr = (CShellAST) ast.getChild(1);
			Object value = exec(expr); // walk/evaluate expr

			// var assign ^('=' a expr)
			ScopeMemorySpace space = getScopeMemorySpace(id.getText());

			// IDֱӸֵΪIDһµID
			space.put(id.getText(), value);
		}
	}
	private Object length(CShellAST ast)
	{
		CShellAST idAST = (CShellAST)ast.getChild(0);
		ScopeMemorySpace space = getScopeMemorySpace(idAST.getText());
		Object id = space.get(idAST.getText());
		if(id == null)
		{
			listener.error("<" + idAST.getText() + ">");
			return null;
		}
		
	
		if(id instanceof String || id instanceof List)
		{
			if(id instanceof String)
			{
				return String.valueOf(id).length();
			}
			else
			{
				List array = (List)id;
				return array.size();
			}
		}
		else
		{
			listener.error("ַֻܼ㳤.");
		}
		return null;
	}
	private void print(CShellAST ast)
	{
		CShellAST expr = (CShellAST) ast.getChild(0);
		Object obj = exec(expr);
		if (obj != null)
			System.out.println(obj);
	}

	private void arrayStat(CShellAST ast)
	{
		List array = new ArrayList();

		CShellAST id = (CShellAST) ast.getChild(0);
		int count = ast.getChildCount() - 1;

		for (int i = 1; i <= count; i++)
		{
			CShellAST valueAST = (CShellAST) ast.getChild(i);
			Object value = exec(valueAST);
			array.add(value);
		}

		// var assign ^('=' a expr)
		ScopeMemorySpace space = getScopeMemorySpace(id.getText());

		// IDֱӸֵΪIDһµID
		space.put(id.getText(), array);

	}

	// ^('if' expr slist elseif_statement_all ^(ELSE else_statement?))
	// ^(ELSEIF ^(expr slist)*);
	private void ifstat(CShellAST ast)
	{
		CShellAST expr = (CShellAST) ast.getChild(0);
		CShellAST ifBlock = (CShellAST) ast.getChild(1);
		CShellAST elseifAll = (CShellAST) ast.getChild(2);
		CShellAST elseBlock = (CShellAST) ast.getChild(3);

		Boolean c = (Boolean) exec(expr);
		if (c.booleanValue())
		{
			exec(ifBlock);
		}
		else
		{
			// ^(ELSEIF ^(expr slist)*);
			if (elseifAll.getChildCount() > 0)
			{
				List<CShellAST> children = elseifAll.getChildren();
				for (CShellAST child : children)
				{
					expr = (CShellAST) child.getChild(0);
					ifBlock = (CShellAST) child.getChild(1);

					c = (Boolean) exec(expr);

					if (c.booleanValue())
					{
						exec(ifBlock);
						return;
					}
				}
			}
			if (elseBlock.getChildCount() == 1)
			{
				exec((CShellAST) elseBlock.getChild(0));
			}
		}

	}

	private boolean lt(CShellAST ast)
	{
		Object a = exec((CShellAST) ast.getChild(0));
		Object b = exec((CShellAST) ast.getChild(1));
		if (a instanceof Number && b instanceof Number)
		{
			Number x = (Number) a;
			Number y = (Number) b;
			return x.floatValue() < y.floatValue();
		}
		return false;
	}

	private boolean gt(CShellAST ast)
	{
		Object a = exec((CShellAST) ast.getChild(0));
		Object b = exec((CShellAST) ast.getChild(1));
		if (a instanceof Number && b instanceof Number)
		{
			Number x = (Number) a;
			Number y = (Number) b;
			return x.floatValue() > y.floatValue();
		}
		return false;
	}

	private Object load(CShellAST ast)
	{
		ScopeMemorySpace space = null;

		if (ast.getType() == CShellParser.ARRAY_ELEMENT)
		{

			space = getScopeMemorySpace(ast.getChild(0).getText());
			if (space != null)
			{
				Object v = exec((CShellAST) ast.getChild(1));

				if (v == null)
				{
					return null;
				}
				int index = Integer.parseInt(String.valueOf(v));
				List array = (List) exec((CShellAST) ast.getChild(0));
				if (array == null)
					return null;

				if (index >= array.size())
				{
					listener.error("Χ<" + index + ">");
					return null;
				}

				else
				{
					return array.get(index);
				}
			}
		}
		else
		{
			space = getScopeMemorySpace(ast.getText());
			if (space != null)
			{
				return space.get(ast.getText());
			}
		}

		listener.error("δҵ " + ast.getText());

		return null;
	}

	private boolean eq(CShellAST ast)
	{
		Object a = exec((CShellAST) ast.getChild(0));
		Object b = exec((CShellAST) ast.getChild(1));
		return a.equals(b);
	}

	public Object add(CShellAST ast)
	{
		Object a = exec((CShellAST) ast.getChild(0));
		Object b = exec((CShellAST) ast.getChild(1));
		if (a instanceof String || b instanceof String)
		{
			return a.toString() + b.toString();
		}
		return op(ast);
	}

	private Object incAndDec(CShellAST ast)
	{

		boolean isArray = (ast.getChild(1).getType() == CShellParser.ARRAY_ELEMENT) ? true
				: false;
		Object idValue = null;
		String idText = null;
		int index = 0;
		int v = 0; // δ仯ֵ
		int v1 = 0; // ѱ仯ֵ
		ScopeMemorySpace space = null;
		List array = null;
		if (isArray)
		{
			idText = ast.getChild(1).getChild(0).getText();
		}
		else
		{
			idText = ast.getChild(1).getText();
		}
		space = getScopeMemorySpace(idText);
		idValue = space.get(idText);
		if (idValue == null)
		{
			listener.error("δҵ<" + idText + ">");
			return null;
		}

		if (isArray)
		{

			array = (List) idValue;
			idValue = exec((CShellAST) ast.getChild(1));
			if (idValue == null)
			{
				return null;
			}
			index = Integer.parseInt(ast.getChild(1).getChild(1).getText());

			v = ((Number) idValue).intValue();
		}
		else
		{
			idValue = exec((CShellAST) ast.getChild(1));
			v = ((Number) idValue).intValue();
		}

		if (ast.getChild(0).getType() == CShellParser.INC)
		{
			if (!(idValue instanceof Integer))
			{
				listener.error("Լӱ.");
			}
			else
			{

				v1 = v + 1;

			}

		}
		else if (ast.getChild(0).getType() == CShellParser.DEC)
		{
			if (!(idValue instanceof Integer))
			{
				listener.error("Լ.");
			}
			else
			{

				v1 = v - 1;

			}
		}
		if (isArray)
		{
			array.set(index, v1);
			space.put(idText, array);
		}
		else
		{
			space.put(idText, v1);
		}
		if (ast.getType() == CShellParser.PREV)
		{
			return v1;
		}
		else if (ast.getType() == CShellParser.SUFFIX)
		{
			return v;
		}

		return null;
	}

	// (INC^ '++' (ID|array_element))
	private Object incAndDec2(CShellAST ast)
	{

		Object id = null;
		int v = 0;
		int v1 = 0;

		if (ast.getType() == CShellParser.INC)
		{
			// ++i
			if (ast.getChild(0).getType() == 1)
			{
				id = exec((CShellAST) ast.getChild(1));
				if (id != null)
				{
					if (!(id instanceof Integer))
					{
						listener.error("Լӱ.");
					}
					else
					{
						v = ((Number) id).intValue();
						v1 = v + 1;

					}
				}
			}

		}
		else if (ast.getChild(0).getType() == CShellParser.DEC)
		{
			if (!(id instanceof Integer))
			{
				listener.error("Լ.");
			}
			else
			{

				v1 = v - 1;

			}
		}
		String s = ast.getChild(1).getText();
		ScopeMemorySpace space = getScopeMemorySpace(s);
		space.put(s, v1);
		if (ast.getChild(0).getType() == 1 || ast.getChild(0).getType() == 2)
		{
			return v1;
		}
		else
		{
			return v;
		}

	}

	public Object dec(CShellAST ast)
	{
		Object id = exec((CShellAST) ast.getChild(0));

		if (id instanceof Integer)
		{
			int v = ((Number) id).intValue();

			v = --v;
			String s = ast.getChild(0).getText();
			ScopeMemorySpace space = getScopeMemorySpace(s);

			// IDֱӸֵΪIDһµID
			space.put(s, v);
			return v;
		}
		listener.error("Լ.");
		return null;
	}

	public Object op(CShellAST ast)
	{

		Object a = exec((CShellAST) ast.getChild(0));
		Object b = exec((CShellAST) ast.getChild(1));
		if (a instanceof Float || b instanceof Float)
		{
			float x = ((Number) a).floatValue();
			float y = ((Number) b).floatValue();
			switch (ast.getType())
			{
				case CShellParser.ADD:
					return x + y;
				case CShellParser.SUB:
					return x - y;
				case CShellParser.MUL:
					return x * y;
				case CShellParser.DIV:
					return x / y;
			}
		}
		if (a instanceof Integer || b instanceof Integer)
		{
			int x = ((Number) a).intValue();
			int y = ((Number) b).intValue();
			switch (ast.getType())
			{
				case CShellParser.ADD:

					return x + y;
				case CShellParser.SUB:
					return x - y;
				case CShellParser.MUL:
					return x * y;
				case CShellParser.DIV:
					return x / y;
			}
		}
		return 0;
	}

	// ^('for' ^(FOR1 for1?) ^(FOR2 for2?) ^(FOR3 for3?) slist)
	private void forloop(CShellAST ast)
	{
		// CShellAST for1 = (CShellAST) ast.getChild(0);
		// System.out.println(for1.getChildCount());
		CShellAST for1 = (CShellAST) ast.getChild(0);
		CShellAST for2 = (CShellAST) ast.getChild(1);
		CShellAST for3 = (CShellAST) ast.getChild(2);
		CShellAST forBlock = (CShellAST) ast.getChild(3);
		ForScopeMemorySpace space = new ForScopeMemorySpace();
		ScopeMemorySpace oldSpace = currentSpace;
		currentSpace = space;
		stack.push(space);
		// CShellAST for3 = (())
		// System.out.println(for1.getChildCount());
		List<CShellAST> children = for1.getChildren();
		if (children != null)
		{
			// forĳʼ
			for (CShellAST child : children)
			{
				exec(child);
			}
		}
		Boolean b = (Boolean) exec((CShellAST) for2.getChild(0));
		for (; b;)
		{
			exec(forBlock);

			children = for3.getChildren();
			if (children != null)
			{
				// forĳʼ
				for (CShellAST child : children)
				{
					exec(child);
				}
			}
			b = (Boolean) exec((CShellAST) for2.getChild(0));
		}

		// System.out.println(forBody.getChildCount());
		// System.out.println(for1.getChild(0).getChild(0).getText());
		/*
		 * PieAST codeStart = (PieAST) t.getChild(1); Boolean c =
		 * (Boolean)exec(condStart); while (c) { exec(codeStart); c = (Boolean)
		 * exec(condStart); }
		 */
		currentSpace = oldSpace;
		stack.pop();
	}
	public void ret(CShellAST ast)
	{
		sharedReturnValue.value = exec((CShellAST) ast.getChild(0));
		throw sharedReturnValue;
	}

	private Object call(CShellAST ast)
	{
		// Resolve function's name
		String fname = ast.getChild(0).getText();
		
		FunctionSymbol fs = (FunctionSymbol) ast.scope.resolve(fname);
		if (fs == null)
		{
			listener.error("δκ<" + fname + ">");
			return null;
		}
		FunctionSpace fspace = new FunctionSpace(fs);
		ScopeMemorySpace oldSpace = currentSpace;
		currentSpace = fspace;
		int argCount = ast.getChildCount() - 1;
		if (fs.formalArgs == null && argCount > 0 || // args compatible?
				fs.formalArgs != null && fs.formalArgs.size() != argCount)
		{
			listener.error("<" + fs.name + ">бȷ.");
			return null;
		}
		int i = 0; // define args according to order in formalArgs
		for (Symbol argS : fs.formalArgs.values())
		{
			VariableSymbol arg = (VariableSymbol) argS;
			CShellAST argAST = (CShellAST) ast.getChild(i + 1);
			Object argValue = exec(argAST);
			fspace.put(arg.name, argValue);
			i++;
		}
		Object result = null;
		stack.push(fspace); 
		//System.out.println(argCount);
		
		try
		{
			exec(fs.blockAST);
		} 
		catch (ReturnValue rv)
		{
			result = rv.value;
		} 
		stack.pop(); // POP arg, locals
		currentSpace = oldSpace;
		return result;
	}
	private ScopeMemorySpace getScopeMemorySpace(String id)
	{

		if (id.startsWith("$"))
		{
			return globals;
		}
		else
		{
			if ("globals".equals(currentSpace.name))
			{

				return globals;
			}
			else
			{
				return stack.peek();
			}

		}

	}

	private ScopeMemorySpace getScopeMemorySpace()
	{
		return globals;
	}

	private void block(CShellAST ast)
	{

		if (ast.getType() != CShellParser.BLOCK)
		{
			listener.error("Ǵ飺" + ast.toStringTree());
		}
		List<CShellAST> stats = ast.getChildren();
		if (stats != null)
		{
			for (int i = 0; i < stats.size(); i++)
			{
				exec(stats.get(i));
			}
		}

	}

}
